/*
 * Adapted from nora.c by Jamey Hicks 9/14/2000
 */

#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>

#include <linux/mtd/mtd.h>
#include <linux/mtd/map.h>


#define WINDOW_ADDR 0xe8000000
#define WINDOW_SIZE 0x02000000

static struct mtd_info *mymtd;

__u8 bitsy_read8(struct map_info *map, unsigned long ofs)
{
  return *(__u8 *)(WINDOW_ADDR + ofs);
}

__u16 bitsy_read16(struct map_info *map, unsigned long ofs)
{
  return *(__u16 *)(WINDOW_ADDR + ofs);
}

__u32 bitsy_read32(struct map_info *map, unsigned long ofs)
{
  return *(__u32 *)(WINDOW_ADDR + ofs);
}

void bitsy_copy_from(struct map_info *map, void *to, unsigned long from, ssize_t len)
{
  memcpy(to, (void *)(WINDOW_ADDR + from), len);
}

void bitsy_write8(struct map_info *map, __u8 d, unsigned long adr)
{
  *(__u8 *)(WINDOW_ADDR + adr) = d;
}

void bitsy_write16(struct map_info *map, __u16 d, unsigned long adr)
{
  *(__u16 *)(WINDOW_ADDR + adr) = d;
}

void bitsy_write32(struct map_info *map, __u32 d, unsigned long adr)
{
  *(__u32 *)(WINDOW_ADDR + adr) = d;
}

void bitsy_copy_to(struct map_info *map, unsigned long to, const void *from, ssize_t len)
{
  memcpy((void *)(WINDOW_ADDR + to), from, len);
}

struct map_info bitsy_map = {
  "BITSY",
  WINDOW_SIZE,
  4,
  bitsy_read8,
  bitsy_read16,
  bitsy_read32,
  bitsy_copy_from,
  bitsy_write8,
  bitsy_write16,
  bitsy_write32,
  bitsy_copy_to,
  0,
  0
};


static int bitsy_mtd_read (struct mtd_info *mtd, loff_t from, size_t len, size_t *retlen, u_char *buf)
{
	return mymtd->read(mymtd, from + (unsigned long)mtd->priv, len, retlen, buf);
}

static int bitsy_mtd_write(struct mtd_info *mtd, loff_t to, size_t len, size_t *retlen, const u_char *buf)
{
	return mymtd->write(mymtd, to + (unsigned long)mtd->priv, len, retlen, buf);
}

static int bitsy_mtd_erase (struct mtd_info *mtd, struct erase_info *instr)
{
	instr->addr += (unsigned long)mtd->priv;
	return mymtd->erase(mymtd, instr);
}

static void bitsy_mtd_sync (struct mtd_info *mtd)
{
	mymtd->sync(mymtd);
}

static int bitsy_mtd_suspend (struct mtd_info *mtd)
{
	return mymtd->suspend(mymtd);
}

static void bitsy_mtd_resume (struct mtd_info *mtd)
{
	mymtd->resume(mymtd);
}


static struct mtd_info bitsy_mtds[10] = {  /* boot, kernel, ramdisk, fs */
	{
		type: MTD_NORFLASH,
		flags: MTD_CAP_NORFLASH,
		size: 0x00040000,
		erasesize: 0x40000,
		name: "BITSY boot firmware",
		module: THIS_MODULE,
		erase: bitsy_mtd_erase,
		read: bitsy_mtd_read,
		write: NULL,
		suspend: NULL,
		resume: NULL,
		sync: NULL,
		priv: (void *)0
	},
	{
		type: MTD_NORFLASH,
		flags: MTD_CAP_NORFLASH,
		size: 0x00080000,
		erasesize: 0x40000,
		name: "BITSY kernel",
		module: THIS_MODULE,
		erase: bitsy_mtd_erase,
		read: bitsy_mtd_read,
		write: bitsy_mtd_write,
		suspend: bitsy_mtd_suspend,
		resume: bitsy_mtd_resume,
		sync: bitsy_mtd_sync,
		priv: (void *)0x40000 /* start of this block */
	},
	{
		type: MTD_NORFLASH,
		flags: MTD_CAP_NORFLASH,
		size: 0x00040000,
		erasesize: 0x40000,
		name: "BITSY params",
		module: THIS_MODULE,
		erase: bitsy_mtd_erase,
		read: bitsy_mtd_read,
		write: bitsy_mtd_write,
		suspend: bitsy_mtd_suspend,
		resume: bitsy_mtd_resume,
		sync: bitsy_mtd_sync,
		priv: (void *)0xC0000 /* start of this block */
	},
	{
		type: MTD_NORFLASH,
		flags: MTD_CAP_NORFLASH,
		size: 0x00100000,
		erasesize: 0x40000,
		name: "BITSY initrd",
		module: THIS_MODULE,
		erase: bitsy_mtd_erase,
		read: bitsy_mtd_read,
		write: bitsy_mtd_write,
		suspend: bitsy_mtd_suspend,
		resume: bitsy_mtd_resume,
		sync: bitsy_mtd_sync,
		priv: (void *)0x00100000 /* start of this block */
	},
	{
		type: MTD_NORFLASH,
		flags: MTD_CAP_NORFLASH,
		size: 0x00300000,
		erasesize: 0x40000,
		name: "BITSY root cramfs",
		module: THIS_MODULE,
		erase: bitsy_mtd_erase,
		read: bitsy_mtd_read,
		write: bitsy_mtd_write,
		suspend: bitsy_mtd_suspend,
		resume: bitsy_mtd_resume,
		sync: bitsy_mtd_sync,
		priv: (void *)0x00200000 /* start of this block */
	},
	{
		type: MTD_NORFLASH,
		flags: MTD_CAP_NORFLASH,
		size: 0x00800000,
		erasesize: 0x40000,
		name: "BITSY usr cramfs",
		module: THIS_MODULE,
		erase: bitsy_mtd_erase,
		read: bitsy_mtd_read,
		write: bitsy_mtd_write,
		suspend: bitsy_mtd_suspend,
		resume: bitsy_mtd_resume,
		sync: bitsy_mtd_sync,
		priv: (void *)0x00500000 /* start of this block */
	},
	{
		type: MTD_NORFLASH,
		flags: MTD_CAP_NORFLASH,
		size: 0x01300000,
		erasesize: 0x40000,
		name: "BITSY usr local",
		module: THIS_MODULE,
		erase: bitsy_mtd_erase,
		read: bitsy_mtd_read,
		write: bitsy_mtd_write,
		suspend: bitsy_mtd_suspend,
		resume: bitsy_mtd_resume,
		sync: bitsy_mtd_sync,
		priv: (void *)0x00d00000 /* start of this block */
	},
};


int __init init_bitsy(void)
{
       	printk(KERN_NOTICE "bitsy flash device: %x at %x\n", WINDOW_SIZE, WINDOW_ADDR);

	mymtd = do_cfi_probe(&bitsy_map);
	if (mymtd) {
                int i;
#ifdef MODULE
		mymtd->module = &__this_module;
#endif
		
                for (i = 0; i < 7; i++)
                   add_mtd_device(&bitsy_mtds[i]);
		return 0;
	}

	return -ENXIO;
}

static void __exit cleanup_bitsy(void)
{
	if (mymtd) {
                int i;
                for (i = 0; i < 7; i++)
                   del_mtd_device(&bitsy_mtds[i]);
		map_destroy(mymtd);
	}
}

module_init(init_bitsy);
module_exit(cleanup_bitsy);
